import torch
import torch.optim as optim
from torch.amp import autocast
from torch.amp import GradScaler
import numpy as np
import pandas as pd
import scipy.sparse
from tqdm import tqdm
from MacSGP.networks import *
from MacSGP.utils import patches, OptimHelper

class Model():

    def __init__(self, adata_st, adata_basis,
                 hidden_dims=[256, 128],
                 n_layers=4,
                 n_SGPs=1,
                 alpha_gcn=None,
                 theta_gcn=None,
                 coef_fe=0.1,
                 coef_reg=1.2,
                 training_steps=3000,
                 lr=0.002,
                 seed=1234,
                 estimate_gamma=False,
                 estimate_gamma_k=True,
                 estimate_alpha=False,
                 ): 
        
        self.training_steps = training_steps

        self.adata_st = adata_st.copy()
        self.celltypes = list(adata_basis.obs.index)

        # add device
        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # set random seed
        torch.manual_seed(seed)
        np.random.seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = True

        self.hidden_dims = [adata_st.shape[1]] + hidden_dims
        self.n_celltype = adata_basis.shape[0]

        G_df = adata_st.uns["Spatial_Net"].copy()
        spots = np.array(adata_st.obs_names)
        spots_id_tran = dict(zip(spots, range(spots.shape[0])))
        G_df["Spot1"] = G_df["Spot1"].map(spots_id_tran)
        G_df["Spot2"] = G_df["Spot2"].map(spots_id_tran)

        G = scipy.sparse.coo_matrix(
            (np.ones(G_df.shape[0]), (G_df["Spot1"], G_df["Spot2"])),
            shape=(adata_st.n_obs, adata_st.n_obs),
        )

        self.edge_index = torch.LongTensor(np.nonzero(G)).to(self.device)

        self.net = SGPNet(hidden_dims=self.hidden_dims,
                            n_layers = n_layers,
                            n_celltypes=self.n_celltype,
                            alpha_gcn = alpha_gcn,
                            theta_gcn = theta_gcn, 
                            estimate_gamma=estimate_gamma,
                            init_gamma=torch.from_numpy(np.array(adata_st.var["gamma"].values)).float().to(self.device),
                            estimate_gamma_k=estimate_gamma_k,
                            estimate_alpha=estimate_alpha,
                            n_SGPs=n_SGPs,
                            coef_fe=coef_fe,
                            coef_reg=coef_reg,
                            ).to(self.device)
        
        self.optimizer = optim.Adamax(list(self.net.parameters()), lr=lr)

        if scipy.sparse.issparse(adata_st.X):
            self.X = torch.from_numpy(adata_st.X.toarray()).float().to(self.device)
        else:
            self.X = torch.from_numpy(adata_st.X).float().to(self.device)
        self.Y = torch.from_numpy(np.array(adata_st.obsm["count"])).float().to(self.device)
        self.lY = torch.from_numpy(np.array(adata_st.obs["library_size"].values.reshape(-1, 1))).float().to(self.device)
        self.basis = torch.from_numpy(np.array(adata_basis.X)).float().to(self.device)
        self.gamma = torch.from_numpy(np.array(adata_st.var["gamma"].values)).float().to(self.device)
        self.alpha = torch.from_numpy(np.array(adata_st.obs["alpha"].values)).float().to(self.device)
        self.proportion = torch.from_numpy(np.array(adata_st.obsm["proportion"])).float().to(self.device)


        self.loss = list()
        self.decon_loss = list()
        self.features_loss = list()
        self.regularization_loss = list()

    def train(
        self,
        report_loss: bool = True,
        step_interval: int = 200,
        test: bool = False,
        gene_patch: bool = False,
        patch_size: int = 200,
        use_amp: bool = False,
    ):
        self.net.train()
        opt = OptimHelper(self.optimizer, use_amp=use_amp)

        for step in tqdm(range(self.training_steps)):
            if gene_patch:
                n_patches = int(np.ceil(self.Y.shape[1] / patch_size))
                opt.zero_grad()
                loss_sum = 0.0
                decon_sum = 0.0
                reg_sum = 0.0

                for start, end in patches(self.Y.shape[1], patch_size):
                    with autocast(enabled=use_amp, device_type = self.device.type):
                        loss_patch = self.net(
                            node_feats=self.X,
                            edge_index=self.edge_index,
                            count_matrix=self.Y[:, start:end],
                            library_size=self.lY,
                            basis=self.basis[:, start:end],
                            alpha=self.alpha,
                            proportion=self.proportion,
                            gene_index=torch.arange(start, end, device=self.device),
                            loss_mode="DECONV",
                            n_patchs=n_patches,  
                            test=test,
                        )
                    loss_sum += float(loss_patch.detach())
                    decon_sum += float(self.net.decon_loss.detach())
                    if hasattr(self.net, "regularization_loss") and self.net.regularization_loss is not None:
                        reg_sum += float(self.net.regularization_loss.detach())
                    opt.backward(loss_patch)

                with autocast(enabled=use_amp, device_type = self.device.type):
                    feature_loss = self.net(
                        node_feats=self.X,
                        edge_index=self.edge_index,
                        count_matrix=None,
                        library_size=None,
                        basis=None,
                        alpha=self.alpha,
                        proportion=self.proportion,
                        gene_index=None,
                        loss_mode="RECON",
                        test=test,
                    )
                loss_sum += float(feature_loss.detach())
                opt.backward(feature_loss)
                opt.step()

                self.loss.append(loss_sum)
                self.decon_loss.append(decon_sum)
                self.features_loss.append(float(self.net.features_loss.detach()))
                self.regularization_loss.append(reg_sum)
            else:
                opt.zero_grad()
                with autocast(enabled=use_amp, device_type = self.device.type):
                    loss = self.net(
                        node_feats=self.X,
                        edge_index=self.edge_index,
                        count_matrix=self.Y,
                        library_size=self.lY,
                        basis=self.basis,
                        alpha=self.alpha,
                        proportion=self.proportion,
                        gene_index=None,
                        test=test,
                    )
                opt.backward(loss)
                opt.step()

                self.loss.append(float(loss.detach()))
                self.decon_loss.append(float(self.net.decon_loss.detach()))
                self.features_loss.append(float(self.net.features_loss.detach()))
                self.regularization_loss.append(
                    float(self.net.regularization_loss.detach()) if hasattr(self.net, "regularization_loss") and self.net.regularization_loss is not None else 0.0
                )

            if report_loss and (step % step_interval == 0):
                tqdm.write(
                    f"Step: {step}, Loss: {self.loss[-1]:.4f}, d_loss: {self.decon_loss[-1]:.4f}, "
                    f"f_loss: {self.features_loss[-1]:.4f}, reg_loss: {self.regularization_loss[-1]:.4f}"
                )

    def eval(self):
        self.net.eval()

        self.Z, self.factor, self.loading, self.gamma, self.alpha_res = self.net.evaluate(self.X, self.edge_index)

        # add learned representations to full ST adata object
        embeddings = self.Z.detach().cpu().numpy()
        cell_reps = pd.DataFrame(embeddings)
        cell_reps.index = self.adata_st.obs.index
        self.adata_st.obsm['latent'] = cell_reps.loc[self.adata_st.obs_names, ].values

        factor = self.factor.detach().cpu().numpy()
        loading = self.loading.detach().cpu().numpy()
        if factor.shape[2] > 1:
            for i in range(factor.shape[0]):
                factor_df = pd.DataFrame(factor[i], index=self.adata_st.obs.index, columns=[f'factor_{j}' for j in range(factor.shape[2])])
                self.adata_st.obsm[f'{self.celltypes[i]}'] = factor_df
            for i in range(loading.shape[0]):
                loading_df = pd.DataFrame(loading[i].T, index=self.adata_st.var.index, columns=[f'loading_{j}' for j in range(loading.shape[1])])
                self.adata_st.varm[f'{self.celltypes[i]}'] = loading_df
        else:
            factor = factor[:, :, 0]
            self.adata_st.obsm['factor'] = pd.DataFrame(factor.T, index=self.adata_st.obs.index, columns=self.celltypes)
            loading = loading[:, 0, :]
            self.adata_st.varm['loading'] = pd.DataFrame(loading.T, index=self.adata_st.var.index, columns=self.celltypes)

        return self.adata_st
    
class Model_deconv():

    def __init__(self, adata_st, adata_basis,
                 hidden_dims=[256, 128],
                 n_layers = 4,
                 alpha_gcn = None,
                 theta_gcn = None, 
                 coef_fe=0.1,
                 training_steps=10000,
                 lr=2e-3,
                 seed=1234,
                 ):
        
        self.training_steps = training_steps

        self.adata_st = adata_st.copy()
        self.celltypes = list(adata_basis.obs.index)

        # add device
        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # set random seed
        torch.manual_seed(seed)
        np.random.seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = True

        self.hidden_dims = [adata_st.shape[1]] + hidden_dims
        self.n_celltype = adata_basis.shape[0]

        G_df = adata_st.uns["Spatial_Net"].copy()
        spots = np.array(adata_st.obs_names)
        spots_id_tran = dict(zip(spots, range(spots.shape[0])))
        G_df["Spot1"] = G_df["Spot1"].map(spots_id_tran)
        G_df["Spot2"] = G_df["Spot2"].map(spots_id_tran)

        G = scipy.sparse.coo_matrix(
            (np.ones(G_df.shape[0]), (G_df["Spot1"], G_df["Spot2"])),
            shape=(adata_st.n_obs, adata_st.n_obs),
        )
        self.edge_index = torch.LongTensor(np.nonzero(G)).to(self.device)

        self.net = DeconvNet(hidden_dims=self.hidden_dims,
                            n_layers = n_layers,
                            n_celltypes=self.n_celltype,
                            alpha_gcn = alpha_gcn,
                            theta_gcn = theta_gcn, 
                            coef_fe=coef_fe,
                            ).to(self.device)
        
        self.optimizer = optim.Adamax(list(self.net.parameters()), lr=lr)

        if scipy.sparse.issparse(adata_st.X):
            self.X = torch.from_numpy(adata_st.X.toarray()).float().to(self.device)
        else:
            self.X = torch.from_numpy(adata_st.X).float().to(self.device)
        self.Y = torch.from_numpy(np.array(adata_st.obsm["count"])).float().to(self.device)
        self.lY = torch.from_numpy(np.array(adata_st.obs["library_size"].values.reshape(-1, 1))).float().to(self.device)
        self.basis = torch.from_numpy(np.array(adata_basis.X)).float().to(self.device)

    def train(
        self,
        report_loss: bool = True,
        step_interval: int = 200,
        test: bool = False,
        gene_patch: bool = False,
        patch_size: int = 200,
        use_amp: bool = False,
    ):
        self.net.train()
        opt = OptimHelper(self.optimizer, use_amp=use_amp)

        for step in tqdm(range(self.training_steps)):
            if gene_patch:
                n_patches = int(np.ceil(self.Y.shape[1] / patch_size))
                opt.zero_grad()
                loss_sum = 0.0
                decon_sum = 0.0

                for start, end in patches(self.Y.shape[1], patch_size):
                    with autocast(enabled=use_amp, device_type = self.device.type):
                        loss_patch = self.net(
                            node_feats=self.X,
                            edge_index=self.edge_index,
                            count_matrix=self.Y[:, start:end],
                            library_size=self.lY,
                            basis=self.basis[:, start:end],
                            gene_index=torch.arange(start, end, device=self.device),
                            loss_mode="DECONV",
                        )
                    loss_sum += float(loss_patch.detach())
                    decon_sum += float(self.net.decon_loss.detach())
                    opt.backward(loss_patch)

                # RECON 一次
                with autocast(enabled=use_amp, device_type = self.device.type):
                    feature_loss = self.net(
                        node_feats=self.X,
                        edge_index=self.edge_index,
                        count_matrix=None,
                        library_size=None,
                        basis=None,
                        gene_index=None,
                        loss_mode="RECON",
                    )
                loss_sum += float(feature_loss.detach())
                opt.backward(feature_loss)
                opt.step()

            else:
                opt.zero_grad()
                with autocast(enabled=use_amp, device_type = self.device.type):
                    loss_sum = self.net(
                        node_feats=self.X,
                        edge_index=self.edge_index,
                        count_matrix=self.Y,
                        library_size=self.lY,
                        basis=self.basis,
                    )
                opt.backward(loss_sum)
                opt.step()
                decon_sum = float(self.net.decon_loss.detach())

            if report_loss and (step % step_interval == 0):
                tqdm.write(
                    f"Step: {step}, Loss: {float(loss_sum):.4f}, d_loss: {float(decon_sum):.4f}, f_loss: {float(self.net.features_loss.detach()):.4f}"
                ) 
    
    def eval(self):
        self.net.eval()
        self.Z, self.beta, self.alpha, self.gamma = self.net.evaluate(self.X, self.edge_index)

        # add learned representations to full ST adata object
        embeddings = self.Z.detach().cpu().numpy()
        cell_reps = pd.DataFrame(embeddings)
        cell_reps.index = self.adata_st.obs.index
        self.adata_st.obsm['latent'] = cell_reps.loc[self.adata_st.obs_names, ].values

        b = self.beta.detach().cpu().numpy()
        alpha = self.alpha.detach().cpu().numpy()
        gamma = self.gamma.detach().cpu().numpy()

        proportion = pd.DataFrame(b, index=self.adata_st.obs.index, columns=self.celltypes)
        self.adata_st.obsm['proportion'] = proportion
        self.adata_st.var['gamma'] = gamma.T
        self.adata_st.obs['alpha'] = alpha

        return self.adata_st



from setuptools import Command, find_packages, setup

__lib_name__ = "MacSGP"
__lib_version__ = "0.1"
__description__ = "MacSGP is a scalable statistical and computational approach for MApping Cell-type-specific Spatial Gene Programs (SGPs) in spatial transcriptomic (ST) data."
__url__ = "https://github.com/YangLabHKUST/MacSGP"
__author__ = "Yeqin Zeng"
__author_email__ = "yzengbj@connect.ust.hk"

setup(
    name = __lib_name__,
    version = __lib_version__,
    description = __description__,
    url = __url__,
    author = __author__,
    author_email = __author_email__,
    zip_safe = False,
    include_package_data = True,
    # long_description=open("README.md").read(),
    # long_description_content_type="text/markdown", 
    license="MIT",
    license_files=("LICENSE",),
    packages=find_packages(),
    install_requires=[
        "anndata==0.11.3",
        "matplotlib==3.10.1",
        "numpy==1.26.4",
        "pandas==2.2.3",
        "scanpy==1.11.0",
        "scipy==1.15.2",
        "torch==2.5.1",
        "torch-geometric==2.6.1",
        "tqdm==4.67.1",
    ],
)


import numpy as np
import scanpy as sc
import anndata as ad
import pandas as pd
import scipy.sparse
import matplotlib
import matplotlib.pyplot as plt
import os
from matplotlib import cm
from matplotlib import colors

def plot_spatial_ct(adata, celltype_plot=None, title = None, show_title=True,
                index='factor', reverse = False,
                proportion_threshold=0., num_threshold=100, 
                cmap='bwr', vmin=None, vmax=None, crop_coords=None,
                spot_size=5, colorbar_loc='right', alpha_img=0.7, hide_image=False, ncols=5,
                return_ax=False, save_path=None, frameon=False
                ):

    celltype = adata.obsm['proportion'].columns
    if celltype_plot is None:
        celltype_plot = celltype
        celltype_plot = list(celltype_plot)
    
    celltype_drop = []
    celltype_filtered = []
    for i in range(len(celltype_plot)):
        if (adata.obsm['proportion'][celltype_plot[i]]>proportion_threshold).sum() < num_threshold:
            celltype_drop.append(celltype_plot[i])
        else:
            celltype_filtered.append(celltype_plot[i])
    print('Dropping cell types:', celltype_drop)
    celltype_plot = celltype_filtered
    
    adata_tmp = adata.copy()
    factor = adata_tmp.obsm[index].copy()

    if title is None:
        title = index

    if len(celltype_plot) == 0:
        print("No cell types to plot after filtering.")
        return
    elif len(celltype_plot) == 1:
        ct = celltype_plot[0]
        adata_tmp.obs['tmp'] = factor[ct]
        filter = adata_tmp.obsm['proportion'][ct]>=proportion_threshold
        if filter.sum() == 0:
            print(f"No spots meet the proportion threshold for cell type {ct}.")
            return
        adata_tmp.obs['tmp'] = adata_tmp.obs['tmp'].where(filter, np.nan)
        if reverse:
            adata_tmp.obs['tmp'] = -adata_tmp.obs['tmp']
        nrows, ncols = 1, 1
        fig, axes = plt.subplots(nrows, ncols, figsize=(5 * ncols, 5 * nrows))
        sc.pl.spatial(
            adata_tmp,
            color='tmp',
            cmap=cmap,
            na_color=None,
            spot_size=spot_size,
            colorbar_loc=colorbar_loc,  
            alpha_img=alpha_img,
            vcenter=0 if index == 'factor' else None,
            vmin=vmin,
            vmax=vmax,
            title=f'{title} of {ct}' if show_title else '',
            frameon=frameon,
            crop_coord=crop_coords,
            ax=axes, 
            show=False,  
            img_key=None if hide_image else 'hires'
        )
        if save_path is not None:
            plt.savefig(save_path)
            plt.close(fig)
            return
        if return_ax:
            return axes
        plt.show()
        return
    
    ncols = ncols
    nrows = (len(celltype_plot) + ncols - 1) // ncols
    fig, axes = plt.subplots(nrows, ncols, figsize=(5 * ncols, 4 * nrows))
    total_slots = nrows * ncols
    
    for i, ct in enumerate(celltype_plot):
        ax = axes.flat[i]
        adata_tmp.obs['tmp'] = factor[ct]
        filter = adata_tmp.obsm['proportion'][ct]>=proportion_threshold
        if filter.sum() == 0:
            print(f"No spots meet the proportion threshold for cell type {ct}.")
        adata_tmp.obs['tmp'] = adata_tmp.obs['tmp'].where(filter, np.nan)

        sc.pl.spatial(
            adata_tmp,
            color='tmp',
            cmap=cmap,
            spot_size=spot_size,
            colorbar_loc=colorbar_loc,
            alpha_img=alpha_img,
            vcenter=0 if index == 'factor' else None,
            vmin=vmin,
            vmax=vmax,
            title=f'{title} of {ct}' if show_title else '',
            frameon=frameon,
            crop_coord=crop_coords,
            ax=ax,  
            show=False,  
            img_key=None if hide_image else 'hires'
        )
    if i + 1 < total_slots:
        for j in range(i + 1, total_slots):
            ax = axes.flat[j]
            ax.set_axis_off()
    
    fig.subplots_adjust(left=0.03, right=0.99, top=0.96, bottom=0.05,
                    wspace=0.08, hspace=0.08)
    if save_path is not None:
        plt.savefig(save_path)
        plt.close(fig)
        return
    if return_ax:
        return axes
    plt.show()



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from torch_geometric.nn.conv import GCN2Conv

class DenseLayer(nn.Module):

    def __init__(self, 
                 nfeat, # dimensionality of input features
                 nout, # dimensionality of output features
                 zero_init=False, # initialize weights as zeros; use Xavier uniform init if zero_init=False
                 ):

        super().__init__()

        self.linear = nn.Linear(nfeat, nout)

        # Initialization
        if zero_init:
            nn.init.zeros_(self.linear.weight.data)
        else:
            nn.init.uniform_(self.linear.weight.data, -np.sqrt(6 / (nfeat + nout)), np.sqrt(6 / (nfeat + nout)))
        nn.init.zeros_(self.linear.bias.data)

    def forward(self, 
            node_feats, # input node features
            ):

        node_feats = self.linear(node_feats)

        return node_feats
    

class GCNII(nn.Module):
    def __init__(self, 
                 nfeat,  # dimensionality of input features
                 nhidden, # dimensionality of hidden features
                 nlayers, # number of GCNII layers
                 alpha=None, # initial residual connection weight
                 theta=None # scaling parameter 
                 ):
        super(GCNII, self).__init__()
        
        self.nlayers = nlayers
        if alpha is None:
            self.alpha = [0.1 for _ in range(nlayers)]
        else:
            self.alpha = alpha
        if theta is None:
            self.theta = [1.0 for _ in range(nlayers)]
        else:
            self.theta = theta
        
        self.fcs = nn.ModuleList()
        self.fcs.append(
            nn.Linear(nfeat, nhidden)
        )

        self.convs = nn.ModuleList()
        for layer in range(nlayers):
            self.convs.append(
                GCN2Conv(nhidden, self.alpha[layer], self.theta[layer], layer+1,
                        add_self_loops=False, normalize=True)
            )
        
        self.act_fn = nn.ELU()

    def forward(self, node_feats, edge_index):
        _layers = []
        node_feats = self.act_fn(self.fcs[0](node_feats))
        _layers.append(node_feats)

        for i,con in enumerate(self.convs):
            node_feats = con(node_feats, _layers[0], edge_index)
            node_feats = self.act_fn(node_feats)
        
        return node_feats
    

class SGPNet(nn.Module):
    
    def __init__(self, 
                hidden_dims, # dimensionality of hidden layers
                n_layers, # number of GCNII layers
                n_celltypes, # number of cell types
                n_SGPs, # number of SGPs for each cell type
                alpha_gcn, # initial residual connection weight for GCNII
                theta_gcn, # scaling parameter for GCNII
                estimate_gamma, # whether to estimate platform effects
                init_gamma, # initial value for platform effects
                estimate_gamma_k, # whether to estimate cell-type-specific unwanted effects
                estimate_alpha, # whether to estimate spot effects
                coef_fe, 
                coef_reg, 
                ):
        
        super().__init__()
        # define layers
        # encoder layers
        self.encoder_layer1 = GCNII(hidden_dims[0],hidden_dims[1],n_layers,alpha_gcn,theta_gcn)
        self.encoder_layer2 = DenseLayer(hidden_dims[1], hidden_dims[2])

        # decoder layers
        self.decoder_layer1 = GCNII(hidden_dims[2],hidden_dims[1],n_layers,alpha_gcn,theta_gcn)
        self.decoder_layer2 = DenseLayer(hidden_dims[1], hidden_dims[0])

        # SGP layers
        #self.SGP_layer = DenseLayer(hidden_dims[2], n_celltypes, zero_init=True)
        self.deconv_factor_layers = nn.ModuleList(
            [DenseLayer(hidden_dims[2], n_SGPs, zero_init=True) for _ in range(n_celltypes)]
        )
        self.loading = nn.Parameter(torch.randn(n_celltypes, n_SGPs, hidden_dims[0]) * 1e-3)

        self.estimate_alpha = estimate_alpha
        if estimate_alpha:
            self.deconv_alpha_layer = DenseLayer(hidden_dims[2], 1, zero_init=True)

        self.estimate_gamma = estimate_gamma
        if estimate_gamma:
            self.gamma = nn.Parameter(init_gamma)
        else:
            self.gamma = init_gamma

        self.estimate_gamma_k = estimate_gamma_k
        if estimate_gamma_k:
            self.gamma_k = nn.Parameter(torch.zeros(n_celltypes, hidden_dims[0]))

        self.coef_fe = coef_fe * 9.25
        self.coef_reg = coef_reg
        self.n_SGPs = n_SGPs

    def forward(self, 
            node_feats, # input node features
            edge_index, # edge index
            count_matrix, # gene expression counts
            library_size, # library size
            basis, # basis matrix
            alpha, # spot-specific effects
            proportion, # cell type proportions
            loss_mode="FULL", # loss mode: "FULL" / "RECON" / "DECONV"
            n_patchs=None, # number of gene patchs
            gene_patch=False, # whether to use gene patch
            gene_index=None, # patch gene index
            patch_size=100, # size of gene patch
            test=False, # test mode
            ):
        # encoder
        Z = self.encoder(node_feats, edge_index)

        if loss_mode in ['FULL', 'RECON']:
            # decoder
            node_feats_recon = self.decoder(Z, edge_index)

            # reconstruction loss of node features
            self.features_loss = torch.mean(torch.sqrt(torch.sum(torch.pow(node_feats-node_feats_recon, 2), axis=1)))

        if loss_mode in ['FULL', 'DECONV']:
            if n_patchs is None:
                n_patchs = 1
            
            # factorization 
            #factor = self.deconv_factor_layer(F.elu(Z))
            factors = [layer(F.elu(Z)) for layer in self.deconv_factor_layers]
            factors = torch.stack(factors)

            if self.estimate_alpha:
                alpha_res = self.deconv_alpha_layer(F.elu(Z))
            else:
                alpha_res = 0
         
            if gene_index is None:
                gene_index = torch.arange(count_matrix.shape[1]).to(count_matrix.device)
            
            # main loss
            if self.estimate_gamma_k:
                basis = basis * torch.exp(self.gamma_k[:, gene_index])
            
            u_exp = torch.exp(torch.matmul(factors, self.loading[:, :, gene_index]))
            u_exp = u_exp * torch.matmul(proportion.T.unsqueeze(2), basis.unsqueeze(1))

            log_lam = torch.log(torch.sum(u_exp, axis=0) + 1e-6) + alpha.unsqueeze(1) + alpha_res + self.gamma.unsqueeze(0)[:, gene_index] 

            self.decon_loss = - torch.mean(torch.sum(count_matrix * 
                                        (torch.log(library_size + 1e-6) + log_lam) - library_size * torch.exp(log_lam), axis=1))
            
            self.regularization_loss = torch.mean(torch.pow(factors, 2)) * proportion.shape[1] / n_patchs+ \
                                torch.sum(torch.pow(self.loading[:, :, gene_index], 2)) / self.gamma.shape[0] / self.n_SGPs

        # total loss
        if loss_mode == 'FULL':
            loss = self.decon_loss + self.coef_fe * self.features_loss + self.coef_reg * self.regularization_loss
        elif loss_mode == 'RECON':
            loss = self.coef_fe * self.features_loss
        elif loss_mode == 'DECONV':
            loss = self.decon_loss + self.coef_reg * self.regularization_loss   

        return loss

    def evaluate(self, node_feats, edge_index):
        # encoder
        Z = self.encoder(node_feats, edge_index)
        
        # factorizer
        #factor = self.deconv_factor_layer(F.elu(Z))
        factors = [layer(F.elu(Z)) for layer in self.deconv_factor_layers]
        factors = torch.stack(factors)
        
        if self.estimate_alpha:
            alpha_res = self.deconv_alpha_layer(F.elu(Z))
        else:
            alpha_res = 0

        return Z, factors, self.loading, self.gamma, alpha_res      
    
    def encoder(self, node_feats, edge_index):
        Z = self.encoder_layer1(node_feats, edge_index)
        Z = self.encoder_layer2(Z)
        return Z        
    
    def decoder(self, Z, edge_index):
        x = self.decoder_layer1(Z, edge_index)
        x = self.decoder_layer2(x)
        return x
    

class DeconvNet(nn.Module):

    def __init__(self, 
                 hidden_dims, # dimensionality of hidden layers
                 n_layers, # number of GCNII layers
                 n_celltypes, # number of cell types
                 alpha_gcn,
                 theta_gcn,
                 coef_fe,
                 ):
        
        super().__init__()
        # define layers
        # encoder layers
        self.encoder_layer1 = GCNII(hidden_dims[0],hidden_dims[1],n_layers,alpha_gcn,theta_gcn)
        self.encoder_layer2 = DenseLayer(hidden_dims[1], hidden_dims[2])

        # decoder layers
        self.decoder_layer1 = GCNII(hidden_dims[2],hidden_dims[1],n_layers,alpha_gcn,theta_gcn)
        self.decoder_layer2 = DenseLayer(hidden_dims[1], hidden_dims[0])
        # deconvolution layers
        self.deconv_alpha_layer = DenseLayer(hidden_dims[2], 1, zero_init=True)
        self.deconv_beta_layer = DenseLayer(hidden_dims[2], n_celltypes, zero_init=True)

        self.gamma = nn.Parameter(torch.zeros(1, hidden_dims[0]))

        self.coef_fe = coef_fe

    def forward(self,
            node_feats, # input node features
            edge_index, # edge index
            count_matrix, # gene expression counts
            library_size, # library size
            basis, # basis matrix
            gene_index=None, # gene index
            loss_mode='FULL', # loss mode: "FULL" / "RECON" / "DECONV"
            ):
        # encoder
        Z = self.encoder(node_feats, edge_index)


        if loss_mode in ['FULL', 'RECON']:
            # decoder
            node_feats_recon = self.decoder(Z, edge_index)

            # reconstruction loss of node features
            self.features_loss = torch.mean(torch.sqrt(torch.sum(torch.pow(node_feats-node_feats_recon, 2), axis=1)))
        
        if loss_mode in ['FULL', 'DECONV']:
            # deconvolutioner
            beta, alpha = self.deconvolutioner(Z)
            if gene_index is None:
                gene_index = torch.arange(count_matrix.shape[1]).to(count_matrix.device)
            # main loss
            log_lam = torch.log(torch.matmul(beta, basis) + 1e-6) + alpha + self.gamma[:, gene_index]
            #lam = torch.exp(log_lam)
            self.decon_loss = - torch.mean(torch.sum(count_matrix * 
                                       (torch.log(library_size + 1e-6) + log_lam) - library_size * torch.exp(log_lam), axis=1))
        
        # Total loss
        if loss_mode == 'FULL':
            loss = self.decon_loss + self.coef_fe * self.features_loss
        elif loss_mode == 'RECON':
            loss = self.coef_fe * self.features_loss
        elif loss_mode == 'DECONV':
            loss = self.decon_loss
        
        return loss
    
    def evaluate(self, node_feats, edge_index):
        # encoder
        Z = self.encoder(node_feats, edge_index)
        
        # deconvolutioner
        beta, alpha = self.deconvolutioner(Z)

        return Z, beta, alpha, self.gamma
    
    def encoder(self, node_feats, edge_index):
        x = self.encoder_layer1(node_feats, edge_index)
        x = self.encoder_layer2(x)
        return x

    def decoder(self, Z, edge_index):
        x = self.decoder_layer1(Z, edge_index)
        x = self.decoder_layer2(x)
        return x   

    def deconvolutioner(self, Z):
        beta = self.deconv_beta_layer(F.elu(Z))
        beta = F.softmax(beta, dim=1)
        H = F.elu(Z)
        alpha = self.deconv_alpha_layer(H)
        return beta, alpha     


import numpy as np
import scanpy as sc
import anndata as ad
import pandas as pd
import scipy.sparse
import torch
from torch_geometric.nn import knn_graph, radius_graph
from torch_geometric.utils import to_undirected
from torch.amp import autocast
from torch.amp import GradScaler

def Cal_Spatial_Net(
    adata,
    mode='Radius', # 'Radius' or 'KNN'
    rad_cutoff=None, # used when mode=='Radius'
    rad_coef=1.1, # used when mode=='Radius' and rad_cutoff is None
    k_cutoff=6, # used when mode=='KNN'
    coor_key='spatial',
    num_workers=1,
    verbose=True,
):
    assert mode in ["Radius", "KNN"]
    if verbose:
        print("Calculating spatial neighbor graph ...")
    if mode == 'KNN':
        edge_index = knn_graph(
            x=torch.tensor(adata.obsm[coor_key]),
            flow="target_to_source",
            k=k_cutoff,
            loop=True,
            num_workers=num_workers,
        )
        edge_index = to_undirected(
            edge_index, num_nodes=adata.shape[0]
        )
    elif mode == 'Radius':
        coords = torch.tensor(adata.obsm[coor_key]).float()
        dist_mat = torch.cdist(coords, coords, p=2)
        eye_mask = torch.eye(coords.size(0), dtype=torch.bool, device=coords.device)
        d_min = dist_mat.masked_fill(eye_mask, float('inf')).min()

        if rad_cutoff is None:
            rad_cutoff = d_min * rad_coef

        edge_index = radius_graph(
            x=coords,
            flow="target_to_source",
            r=rad_cutoff,
            loop=True,  
            num_workers=num_workers, 
        )

    graph_df = pd.DataFrame(edge_index.numpy().T, columns=["Spot1", "Spot2"])
    id_spot_trans = dict(zip(range(adata.n_obs), adata.obs_names))
    graph_df["Spot1"] = graph_df["Spot1"].map(id_spot_trans)
    graph_df["Spot2"] = graph_df["Spot2"].map(id_spot_trans)
    adata.uns["Spatial_Net"] = graph_df

    if verbose:
        print(f"The graph contains {graph_df.shape[0]} edges, {adata.n_obs} spots.")
        print(f"{graph_df.shape[0]/adata.n_obs} neighbors per spot on average.")

def preprocess(
    adata_st_input,
    adata_ref_input,
    celltype_ref_col='celltype',
    sample_col=None,
    celltype_ref=None,
    n_hvg_group=500,
    coor_key="spatial",
    min_genes=1,
    min_cells=1,
):
    adata_st = adata_st_input.copy()

    print("Finding highly variable genes...")
    adata_ref = adata_ref_input.copy()
    adata_ref.var_names_make_unique()
    # Remove mt-genes
    adata_ref = adata_ref[
        :,
        np.array(~adata_ref.var.index.isna())
        & np.array(~adata_ref.var_names.str.startswith("mt-"))
        & np.array(~adata_ref.var_names.str.startswith("MT-")),
    ]
    if celltype_ref is not None:
        if not isinstance(celltype_ref, list):
            raise ValueError("'celltype_ref' must be a list!")
        else:
            adata_ref = adata_ref[
                [
                    (t in celltype_ref)
                    for t in adata_ref.obs[celltype_ref_col].values.astype(str)
                ],
                :,
            ]
    else:
        celltype_counts = adata_ref.obs[celltype_ref_col].value_counts()
        celltype_ref = list(celltype_counts.index[celltype_counts > 1])
        adata_ref = adata_ref[
            [
                (t in celltype_ref)
                for t in adata_ref.obs[celltype_ref_col].values.astype(str)
            ],
            :,
        ]

    # Remove cells and genes with zero counts
    sc.pp.filter_cells(adata_ref, min_genes=min_genes)
    sc.pp.filter_genes(adata_ref, min_cells=min_cells)
    
    # preprocess ST adatas
    adata_st.var_names_make_unique()
    # Remove mt-genes
    adata_st = adata_st[
        :,
        (
            np.array(~adata_st.var.index.str.startswith("mt-"))
            & np.array(~adata_st.var.index.str.startswith("MT-"))
        ),
    ]

    # Take gene intersection
    #genes = list(adata_st.var.index & adata_ref.var.index)
    genes = list(adata_st.var_names.intersection(adata_ref.var_names))
    adata_ref = adata_ref[:, genes]
    adata_st = adata_st[:, genes]

    # Select hvgs
    adata_ref_log = adata_ref.copy()
    sc.pp.log1p(adata_ref_log)
    hvgs = select_hvgs(
        adata_ref_log, celltype_ref_col=celltype_ref_col, num_per_group=n_hvg_group
    )

    print("%d highly variable genes selected." % len(hvgs))
    adata_ref = adata_ref[:, hvgs]

    print("Calculate basis for deconvolution...")
    sc.pp.filter_cells(adata_ref, min_genes=1)
    sc.pp.normalize_total(adata_ref, target_sum=1)
    celltype_list = list(
        sorted(set(adata_ref.obs[celltype_ref_col].values.astype(str)))
    )

    basis = np.zeros((len(celltype_list), len(adata_ref.var.index)))
    if sample_col is not None:
        sample_list = list(sorted(set(adata_ref.obs[sample_col].values.astype(str))))
        for i in range(len(celltype_list)):
            c = celltype_list[i]
            tmp_list = []
            for j in range(len(sample_list)):
                s = sample_list[j]
                tmp = adata_ref[
                    (adata_ref.obs[celltype_ref_col].values.astype(str) == c)
                    & (adata_ref.obs[sample_col].values.astype(str) == s),
                    :,
                ].X
                if scipy.sparse.issparse(tmp):
                    tmp = tmp.toarray()
                if tmp.shape[0] >= 3:
                    tmp_list.append(np.mean(tmp, axis=0).reshape((-1)))
            tmp_mean = np.mean(tmp_list, axis=0)
            if scipy.sparse.issparse(tmp_mean):
                tmp_mean = tmp_mean.toarray()
            print(
                "%d batches are used for computing the basis vector of cell type <%s>."
                % (len(tmp_list), c)
            )
            basis[i, :] = tmp_mean
    else:
        for i in range(len(celltype_list)):
            c = celltype_list[i]
            tmp = adata_ref[
                adata_ref.obs[celltype_ref_col].values.astype(str) == c, :
            ].X
            if scipy.sparse.issparse(tmp):
                tmp = tmp.toarray()
            basis[i, :] = np.mean(tmp, axis=0).reshape((-1))
    
    adata_basis = ad.AnnData(X=basis)
    df_gene = pd.DataFrame({"gene": adata_ref.var.index})
    df_gene = df_gene.set_index("gene")
    df_celltype = pd.DataFrame({"celltype": celltype_list})
    df_celltype = df_celltype.set_index("celltype")
    adata_basis.obs = df_celltype
    adata_basis.var = df_gene
    adata_basis = adata_basis[~np.isnan(adata_basis.X[:, 0])]

    print("Preprocess ST data...")
    # Store counts and library sizes for Poisson modeling
    st_mtx = adata_st[:, hvgs].X.copy()
    if scipy.sparse.issparse(st_mtx):
        st_mtx = st_mtx.toarray()
    adata_st.obsm["count"] = st_mtx
    st_library_size = np.sum(st_mtx, axis=1)
    adata_st.obs["library_size"] = st_library_size

    # Normalize ST data
    sc.pp.normalize_total(adata_st, target_sum=1e4)
    sc.pp.log1p(adata_st)
    adata_st = adata_st[:, hvgs]
    if scipy.sparse.issparse(adata_st.X):
        adata_st.X = adata_st.X.toarray()

    return adata_st, adata_basis


def select_hvgs(adata_ref, celltype_ref_col, num_per_group=200):
    sc.tl.rank_genes_groups(
        adata_ref,
        groupby=celltype_ref_col,
        method="t-test",
        key_added="ttest",
        use_raw=False,
    )
    markers_df = pd.DataFrame(adata_ref.uns["ttest"]["names"]).iloc[0:num_per_group, :]
    genes = sorted(list(np.unique(markers_df.melt().value.values)))
    return genes


class OptimHelper:
    def __init__(self, optimizer, use_amp: bool):
        self.optimizer = optimizer
        self.scaler = GradScaler(enabled=use_amp)
        self.use_amp = use_amp

    def zero_grad(self):
        self.optimizer.zero_grad(set_to_none=True)

    def backward(self, loss):
        if self.use_amp:
            self.scaler.scale(loss).backward()
        else:
            loss.backward()

    def step(self):
        if self.use_amp:
            self.scaler.step(self.optimizer)
            self.scaler.update()
        else:
            self.optimizer.step()

def patches(total_cols: int, size: int):
    for s in range(0, total_cols, size):
        e = min(s + size, total_cols)
        yield s, e